@@ -10,8 +10,8 @@ use std::{
 use base64::{decode_config, encode_config, Config, STANDARD_NO_PAD, URL_SAFE_NO_PAD};
 use ed25519_dalek::Digest;
 use ruma_identifiers::{EventId, RoomVersionId, ServerNameBox, UserId};
-use ruma_serde::{to_canonical_json_string, CanonicalJsonObject, CanonicalJsonValue};
-use serde_json::from_str as from_json_str;
+use ruma_serde::{CanonicalJsonObject, CanonicalJsonValue};
+use serde_json::{from_str as from_json_str, to_string as to_json_string};
 use sha2::Sha256;
 
 use crate::{
@@ -21,6 +21,8 @@ use crate::{
     Error, JsonError, JsonType, ParseError, VerificationError,
 };
 
+const MAX_PDU_BYTES: usize = 65_535;
+
 /// The fields that are allowed to remain in an event during redaction.
 static ALLOWED_KEYS: &[&str] = &[
     "event_id",
@@ -149,7 +151,7 @@ where
     let maybe_unsigned_entry = object.remove_entry("unsigned");
 
     // Get the canonical JSON string.
-    let json = to_canonical_json_string(object).map_err(JsonError::CanonicalJson)?;
+    let json = to_json_string(object).map_err(JsonError::Serde)?;
 
     // Sign the canonical JSON string.
     let signature = key_pair.sign(json.as_bytes());
@@ -194,20 +196,24 @@ where
 ///     }"#;
 ///
 /// let object = serde_json::from_str(input).unwrap();
-/// let canonical = ruma_signatures::canonical_json(&object);
+/// let canonical = ruma_signatures::canonical_json(&object).unwrap();
 ///
 /// assert_eq!(canonical, r#"{"日":1,"本":2}"#);
 /// ```
-pub fn canonical_json(object: &CanonicalJsonObject) -> String {
+pub fn canonical_json(object: &CanonicalJsonObject) -> Result<String, Error> {
     canonical_json_with_fields_to_remove(object, CANONICAL_JSON_FIELDS_TO_REMOVE)
 }
 
 /// Uses a set of public keys to verify a signed JSON object.
 ///
+/// Unlike `content_hash` and `reference_hash`, this function does not report an error if the
+/// canonical JSON is larger than 65535 bytes; this function may be used for requests that are
+/// larger than just one PDU's maximum size.
+///
 /// # Parameters
 ///
 /// * public_key_map: A map from entity identifiers to a map from key identifiers to public keys.
-/// Generally, entity identifiers are server names—the host/IP/port of a homeserver (e.g.
+/// Generally, entity identifiers are server names — the host/IP/port of a homeserver (e.g.
 /// "example.com") for which a signature must be verified. Key identifiers for each server (e.g.
 /// "ed25519:1") then map to their respective public keys.
 /// * object: The JSON object that was signed.
@@ -321,7 +327,7 @@ fn verify_json_with<V>(
 where
     V: Verifier,
 {
-    verifier.verify_json(public_key, signature, canonical_json(object).as_bytes())
+    verifier.verify_json(public_key, signature, canonical_json(object)?.as_bytes())
 }
 
 /// Creates a *content hash* for an event.
@@ -334,11 +340,19 @@ where
 /// # Parameters
 ///
 /// object: A JSON object to generate a content hash for.
-pub fn content_hash(object: &CanonicalJsonObject) -> String {
-    let json = canonical_json_with_fields_to_remove(object, CONTENT_HASH_FIELDS_TO_REMOVE);
+///
+/// # Errors
+///
+/// Returns an error if the event is too large.
+pub fn content_hash(object: &CanonicalJsonObject) -> Result<String, Error> {
+    let json = canonical_json_with_fields_to_remove(object, CONTENT_HASH_FIELDS_TO_REMOVE)?;
+    if json.len() > MAX_PDU_BYTES {
+        return Err(Error::PduSize);
+    }
+
     let hash = Sha256::digest(json.as_bytes());
 
-    encode_config(&hash, STANDARD_NO_PAD)
+    Ok(encode_config(&hash, STANDARD_NO_PAD))
 }
 
 /// Creates a *reference hash* for an event.
@@ -355,7 +369,7 @@ pub fn content_hash(object: &CanonicalJsonObject) -> String {
 ///
 /// # Errors
 ///
-/// Returns an error if redaction fails.
+/// Returns an error if the event is too large or redaction fails.
 pub fn reference_hash(
     value: &CanonicalJsonObject,
     version: &RoomVersionId,
@@ -363,7 +377,10 @@ pub fn reference_hash(
     let redacted_value = redact(value, version)?;
 
     let json =
-        canonical_json_with_fields_to_remove(&redacted_value, REFERENCE_HASH_FIELDS_TO_REMOVE);
+        canonical_json_with_fields_to_remove(&redacted_value, REFERENCE_HASH_FIELDS_TO_REMOVE)?;
+    if json.len() > MAX_PDU_BYTES {
+        return Err(Error::PduSize);
+    }
 
     let hash = Sha256::digest(json.as_bytes());
 
@@ -481,7 +498,7 @@ pub fn hash_and_sign_event<K>(
 where
     K: KeyPair,
 {
-    let hash = content_hash(object);
+    let hash = content_hash(object)?;
 
     let hashes_value = object
         .entry("hashes".to_owned())
@@ -598,7 +615,7 @@ pub fn verify_event(
     };
 
     let servers_to_check = servers_to_check_signatures(object, version)?;
-    let canonical_json = from_json_str(&canonical_json(&redacted)).map_err(JsonError::from)?;
+    let canonical_json = from_json_str(&canonical_json(&redacted)?).map_err(JsonError::from)?;
 
     for entity_id in servers_to_check {
         let signature_set = match signature_map.get(entity_id.as_str()) {
@@ -657,7 +674,7 @@ pub fn verify_event(
         verify(STANDARD_NO_PAD)?;
     }
 
-    let calculated_hash = content_hash(object);
+    let calculated_hash = content_hash(object)?;
 
     if *hash == calculated_hash {
         Ok(Verified::All)
@@ -671,16 +688,20 @@ struct SignatureAndPubkey<'a> {
     public_key: &'a String,
 }
 
-/// Internal implementation detail of the canonical JSON algorithm. Allows customization of the
-/// fields that will be removed before serializing.
-fn canonical_json_with_fields_to_remove(object: &CanonicalJsonObject, fields: &[&str]) -> String {
+/// Internal implementation detail of the canonical JSON algorithm.
+///
+/// Allows customization of the fields that will be removed before serializing.
+fn canonical_json_with_fields_to_remove(
+    object: &CanonicalJsonObject,
+    fields: &[&str],
+) -> Result<String, Error> {
     let mut owned_object = object.clone();
 
     for field in fields {
         owned_object.remove(*field);
     }
 
-    to_canonical_json_string(&owned_object).expect("JSON object serialization to succeed")
+    to_json_string(&owned_object).map_err(|e| Error::Json(e.into()))
 }
 
 /// Redacts an event using the rules specified in the Matrix client-server specification.
@@ -879,7 +900,7 @@ mod tests {
             _ => unreachable!(),
         };
 
-        assert_eq!(canonical_json(&object), canonical);
+        assert_eq!(canonical_json(&object).unwrap(), canonical);
     }
 
     #[test]
